from datetime import date
import json
from typing import Optional

from ssb_altinn3_util.models.skjemadata.skjemadata_request_models import SkjemaRequestModel

from dapla_suv_tools._internals.integration import api_client
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util.validators import skjema_id_validator, ra_nummer_validator


@SuvOperationContext(validator=skjema_id_validator)
def get_skjema_by_id(*, skjema_id: int, context: SuvOperationContext = None) -> OperationResult:
    try:
        content: str = api_client._get(path=f"/skjemadata/skjema/{skjema_id}")
        content_json = json.loads(content)
        return OperationResult(success=True, value=content_json)

    except Exception as e:
        context.set_error(type(e).__name__, f"Failed to fetch for id {skjema_id}", e)
        return OperationResult(success=False, value=context.errors())


@SuvOperationContext(validator=ra_nummer_validator)
def get_skjema_by_ra_nummer(*, ra_nummer: str, max_results: int = 10, latest_only: bool = False, context: SuvOperationContext = None) -> dict:
    try:
        filters = json.dumps({"ra_nummer": ra_nummer})
        content: str = api_client._post(
            path=f"/skjemadata/skjema_paged?size={max_results}&order_by=versjon&asc=false", body_json=filters
        )

        result: dict = json.loads(content)

        if latest_only:
            return result["results"][0]

        return result["results"]

    except Exception as e:
        raise Exception(f"Failed to fetch for ra_nummer '{ra_nummer}'.", e)


@SuvOperationContext(validator=ra_nummer_validator)
def create_skjema(
        *,
        ra_nummer: str,
        versjon: int,
        undersokelse_nr: str,
        gyldig_fra: date,
        endret_av: str,
        datamodell: Optional[str] = None,
        beskrivelse: Optional[str] = None,
        navn_nb: Optional[str] = None,
        navn_nn: Optional[str] = None,
        navn_en: Optional[str] = None,
        infoside: Optional[str] = None,
        eier: Optional[str] = None,
        kun_sky: bool = False,
        gyldig_til: Optional[date] = None,
        context: SuvOperationContext = None
) -> int:
    model = SkjemaRequestModel(
        ra_nummer=ra_nummer,
        versjon=versjon,
        undersokelse_nr=undersokelse_nr,
        gyldig_fra=gyldig_fra,
        gyldig_til=gyldig_til,
        endret_av=endret_av,
        datamodell=datamodell,
        beskrivelse=beskrivelse,
        navn_nb=navn_nb,
        navn_nn=navn_nn,
        navn_en=navn_en,
        infoside=infoside,
        eier=eier,
        kun_sky="J" if kun_sky else "N"
    )

    try:
        body = model.model_dump_json()
        content: str = api_client._post(path=f"/skjemadata/skjema", body_json=body)
        return json.loads(content)["id"]
    except Exception as e:
        raise Exception(f"Failed to create for ra_number '{ra_nummer}' - version '{versjon}'", e)


@SuvOperationContext(validator=skjema_id_validator)
def delete_skjema(*, skjema_id: int, context: SuvOperationContext = None) -> bool:
    pass



from datetime import date
import json
from typing import Optional

from dapla_suv_tools._internals.integration import api_client as api_client
from ssb_altinn3_util.models.skjemadata.skjemadata_request_models import SkjemaRequestModel

from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util.validators import skjema_id_validator, ra_nummer_validator


def get_skjema_by_id(*, skjema_id: int) -> dict:
    with SuvOperationContext(validator=skjema_id_validator, func_kwargs=locals()) as s:
        try:
            content: str = api_client._get(path=f"/skjemadata/skjema/{skjema_id}")
            return json.loads(content)
        except Exception as e:
            raise Exception(f"Failed to fetch for id {skjema_id}", e)


def get_skjema_by_ra_nummer(*, ra_nummer: str, max_results: int = 10, latest_only: bool = False) -> dict:

    with SuvOperationContext(validator=ra_nummer_validator, func_kwargs=locals()) as s:
        try:
            filters = json.dumps({"ra_nummer": ra_nummer})
            content: str = api_client._post(
                path=f"/skjemadata/skjema_paged?size={max_results}&order_by=versjon&asc=false", body_json=filters
            )

            result: dict = json.loads(content)

            if latest_only:
                return result["results"][0]

            return result["results"]

        except Exception as e:
            raise Exception(f"Failed to fetch for ra_nummer '{ra_nummer}'.", e)


def create_skjema(
        *,
        ra_nummer: str,
        versjon: int,
        undersokelse_nr: str,
        gyldig_fra: date,
        endret_av: str,
        datamodell: Optional[str] = None,
        beskrivelse: Optional[str] = None,
        navn_nb: Optional[str] = None,
        navn_nn: Optional[str] = None,
        navn_en: Optional[str] = None,
        infoside: Optional[str] = None,
        eier: Optional[str] = None,
        kun_sky: bool = False,
        gyldig_til: Optional[date] = None,
) -> int:
    model = SkjemaRequestModel(
        ra_nummer=ra_nummer,
        versjon=versjon,
        undersokelse_nr=undersokelse_nr,
        gyldig_fra=gyldig_fra,
        gyldig_til=gyldig_til,
        endret_av=endret_av,
        datamodell=datamodell,
        beskrivelse=beskrivelse,
        navn_nb=navn_nb,
        navn_nn=navn_nn,
        navn_en=navn_en,
        infoside=infoside,
        eier=eier,
        kun_sky="J" if kun_sky else "N"
    )

    try:
        body = model.model_dump_json()
        content: str = api_client._post(path=f"/skjemadata/skjema", body_json=body)
        return json.loads(content)["id"]
    except Exception as e:
        raise Exception(f"Failed to create for ra_number '{ra_nummer}' - version '{versjon}'", e)


def delete_skjema(*, skjema_id: int) -> bool:
    pass


def get_periode_by_id(*, periode_id: int) -> dict:
    try:
        content: str = api_client._get(path=f"/skjemadata/periode/{periode_id}")
        return json.loads(content)
    except Exception as e:
        raise Exception(f"Failed to fetch for id {periode_id}", e)


def get_perioder_by_skjema_id(*, skjema_id: int) -> dict:
    try:
        content: str = api_client._get(path=f"/skjemadata/periode/skjema/{skjema_id}")
        return json.loads(content)
    except Exception as e:
        raise Exception(f"Failed to fetch for id {skjema_id}", e)



from functools import wraps
from typing import Any, Callable, Optional
import inspect


class SuvOperationContext:

    state: dict[str, Any]
    func_kwargs: dict[str, Any]
    arg_validator: Callable[["SuvOperationContext", dict], None]

    def __init__(self, validator: Callable[["SuvOperationContext", dict], None] = None, func_kwargs: dict[str, Any] = None):
        self.arg_validator = validator
        self.state = {
            "log": [],
            "errors": []
        }
        self.func_kwargs = func_kwargs

    def __enter__(self):
        try:
            if self.func_kwargs and self.arg_validator:
                self.arg_validator(self, **self.func_kwargs)

            return self
        except Exception as e:
            self.__exit__(type(e), e, e.__traceback__)
            raise

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        if exc_type is not None:
            if self.state["errors"]:
                print("Errors flagged during operation:")
                for error in self.state["errors"]:
                    print(error)
            return False
        return True

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if self.arg_validator is not None:
                self.arg_validator(**kwargs)
            self.func_args = args
            self.func_kwargs = kwargs
            with self._recreate_context_manager():
                sig = inspect.signature(func)
                if "context" in sig.parameters or "kwargs" in sig.parameters:
                    return func(*args, **kwargs, context=self)
                return func(*args, **kwargs)

        return wrapper

    def _recreate_context_manager(self):
        return self

    def log(self, level: str, operation: str, message: str = "", result: str = "OK"):
        self.state["log"].append({
            "level": level,
            "operation": operation,
            "message": message,
            "result": result
        })

    def set_error(self, error_type: str, error_msg: str, exception: Exception):
        self.state["errors"].append({
            "error_type": error_type,
            "error_message": error_msg,
            "exception": exception
        })

    def errors(self) -> dict:
        return {"errors": self.state["errors"]}


###########  REMOVE!

# @SuvOperationContext()
# def try_empty_args_1():
#     print("Hello from mister empty 1!")
#
#
# @SuvOperationContext(stuff="STUFF")
# def try_empty_args_2():
#     print("Hello from mister empty 2!")
#
#
# @SuvOperationContext()
# def try_func_args_1(a):
#     print("Hello from func 1", a)


def bob_validator(**kwargs):
    if "bob" not in kwargs:
        raise ValueError("No bob here!")
    if len(kwargs["bob"]) < 5:
        raise ValueError(f"Not enough bob!  (bob was '{kwargs['bob']}')")


@SuvOperationContext(validator=bob_validator)
def try_func_args_2(a, bob, **kwargs):
    print("Hello from func 2", a, bob)
    if "context" in kwargs:
        kwargs["context"].state["visited"] = "func 2"


@SuvOperationContext(validator=bob_validator)
def try_func_args_3(a, bob, context: SuvOperationContext):
    print("Hello from func 3", a, bob)
    print(context.state)
    context.log("info", "visit", "visited func 3")
    try:
        b = 10 / a
        print(f"10 / {a} = {b}")
    except Exception as e:
        context.set_error(
            type(e).__name__,
            f"An error occurred dividing 10 with {a}",
            e
        )
        raise


# def try_func_args_4(a, bob):
#     with SuvOperationContext(validator=bob_validator, params=locals()) as s:
#         print("Hello from func 4", a, bob)

#try_empty_args_1()
#try_empty_args_2()
#try_func_args_1(1)
#try_func_args_2(2, bob="Snobb")
#try_func_args_3(0, bob="Snobb")
# try_func_args_3(3, bob="Globs")
#try_func_args_4(a=5, bob="Tobb")



import json
import os
import requests

from dapla.auth import AuthClient

END_USER_API_BASE_URL = os.getenv("SUV_END_USER_API_URL")


def _get(path: str) -> str:
    headers = _get_headers()

    response = requests.get(f"{END_USER_API_BASE_URL}{path}", headers=headers)

    return _handle_response(response=response)


def _post(path: str, body_json: str) -> str:
    headers = _get_headers()

    response = requests.post(url=f"{END_USER_API_BASE_URL}{path}", headers=headers, data=body_json)

    return _handle_response(response=response)


def _handle_response(response: requests.Response) -> str:

    if not _success(response.status_code):
        error = response.content.decode("UTF-8")
        print(f"Error (status: {response.status_code}):  {error}")
        raise Exception("Failed to fetch.")

    return response.content.decode("UTF-8")


def _get_headers() -> dict:
    token: str = AuthClient.fetch_personal_token()

    return {
        "authorization": f"Bearer {token}",
        "content-type": "application/json"
    }


def _success(status_code: int) -> bool:
    return str(status_code).startswith("2")
